#!/usr/bin/env python

"""
Managing and presenting periods of time.

Copyright (C) 2014, 2015 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from bisect import bisect_left, insort_left
from imiptools.dates import get_datetime, get_start_of_day, to_timezone

# Time management.

def have_conflict(freebusy, periods, get_conflicts=False):

    """
    Return whether any period in 'freebusy' overlaps with the given 'periods',
    returning a collection of such overlapping periods if 'get_conflicts' is
    set to a true value.
    """

    conflicts = []
    for start, end in periods:
        overlapping = period_overlaps(freebusy, (start, end), get_conflicts)
        if overlapping:
            if get_conflicts:
                conflicts += overlapping
            else:
                return True

    if get_conflicts:
        return conflicts
    else:
        return False

def insert_period(freebusy, period):
    insort_left(freebusy, period)

def remove_period(freebusy, uid):
    i = 0
    while i < len(freebusy):
        t = freebusy[i]
        if len(t) >= 3 and t[2] == uid:
            del freebusy[i]
        else:
            i += 1

def period_overlaps(freebusy, period, get_periods=False):

    """
    Return whether any period in 'freebusy' overlaps with the given 'period',
    returning a collection of overlapping periods if 'get_periods' is set to a
    true value.
    """

    dtstart, dtend = period[:2]
    found = bisect_left(freebusy, (dtstart, dtend, None, None))

    overlapping = []

    # Find earlier overlapping periods.

    i = found

    while i > 0 and freebusy[i - 1][1] > dtstart:
        if get_periods:
            overlapping.insert(0, freebusy[i - 1])
        else:
            return True
        i -= 1

    # Find later overlapping periods.

    i = found

    while i < len(freebusy) and (dtend is None or freebusy[i][0] < dtend):
        if get_periods:
            overlapping.append(freebusy[i])
        else:
            return True
        i += 1

    if get_periods:
        return overlapping
    else:
        return False

# Period layout.

def get_scale(l):

    """
    Return an ordered time scale from the given list 'l' of tuples, with the
    first two elements of each tuple being start and end times.

    The returned scale is a collection of (time, (starting, ending)) tuples,
    where starting and ending are collections of tuples from 'l'.
    """

    scale = {}

    for t in l:
        start, end = t[:2]

        # Add a point and this event to the starting list.

        if not scale.has_key(start):
            scale[start] = [], []
        scale[start][0].append(t)

        # Add a point and this event to the ending list.

        if not scale.has_key(end):
            scale[end] = [], []
        scale[end][1].append(t)

    scale = scale.items()
    scale.sort()
    return scale

def get_slots(l):

    """
    Return an ordered list of time slots from the given list 'l' of tuples, with
    the first two elements of each tuple being start and end times.

    Each slot is a tuple containing a point in time for the start of the slot,
    together with a list of parallel event tuples, each tuple containing the
    original details of an event.
    """

    slots = []
    active = []

    for point, (starting, ending) in get_scale(l):

        # Discard all active events ending at or before this start time.

        for t in ending:
            i = active.index(t)
            active[i] = None

        for t in starting:
            try:
                i = active.index(None)
                active[i] = t
            except ValueError:
                active.append(t)

        while active and active[-1] is None:
            active.pop()

        slots.append((point, active[:]))

    return slots

def partition_slots(slots, tzid):

    """
    Partition the given 'slots' into separate collections having a date-level
    resolution, using the given 'tzid' to make sure that the day boundaries are
    defined according to the chosen time zone.

    Return a collection of (date, slots) tuples.
    """

    partitioned = {}
    current = None
    current_date = None
    previously_active = None

    for point, active in slots:
        dt = to_timezone(get_datetime(point), tzid)
        start_of_day = get_start_of_day(dt)
        this_date = dt.date()

        # For each new day, create a partition of the original collection.

        if this_date != current_date:
            current_date = this_date
            partitioned[current_date] = current = []

            # Add any continuing periods.

            if dt != start_of_day and previously_active:
                current.append((start_of_day, previously_active))

        # Add the currently active periods at this point in time.

        current.append((point, active))
        previously_active = active

    partitioned = partitioned.items()
    partitioned.sort()
    return partitioned

def get_spans(slots):

    "Inspect the given 'slots', returning a mapping of event uids to spans."

    points = [point for point, active in slots]
    spans = {}

    for point, active in slots:
        for t in active:
            if t:
                start, end, uid = t[:3]
                try:
                    start_slot = points.index(start)
                except ValueError:
                    start_slot = 0
                try:
                    end_slot = points.index(end)
                except ValueError:
                    end_slot = len(slots)
                spans[uid] = end_slot - start_slot

    return spans

# vim: tabstop=4 expandtab shiftwidth=4
